"""Support for the Dynalite devices."""

from .const import DOMAIN


class DynaliteBaseDevice:  # Deriving from Object so it doesn't override the entity (light, switch, device, etc.)
    """Base class for Dynalite devices."""

    def __init__(self, area, area_name, name, master_area, bridge):
        """Initialize the device."""
        self._area = area
        self._area_name = area_name
        self._name = name
        self._master_area = master_area
        self._bridge = bridge
        self._listeners = []

    @property
    def name(self):
        """Return the name of the device."""
        return self._name

    @property
    def area_name(self):
        """Return the name of the area."""
        return self._area_name

    @property
    def available(self):
        """Return if device is available."""
        return self._bridge.available

    @property
    def get_master_area(self):
        """Get the master area when combining entities from different Dynet areas to the same area."""
        return self._master_area

    def add_listener(self, listener):
        """Add a listener for changes."""
        self._listeners.append(listener)

    def update_listeners(self, stop_fade=False):
        """Update all listeners."""
        for listener in self._listeners:
            listener(self, stop_fade)


class DynaliteChannelBaseDevice(DynaliteBaseDevice):
    """Representation of a Dynalite Channel as a Home Assistant device."""

    def __init__(
        self, area, area_name, channel, name, type, master_area, bridge, device
    ):
        """Initialize the device."""
        self._channel = channel
        self._type = type
        self._device = device
        super().__init__(area, area_name, name, master_area, bridge)

    @property
    def unique_id(self):
        """Return the ID of this device."""
        return "dynalite_area_" + str(self._area) + "_channel_" + str(self._channel)

    def stop_fade(self):
        """Update the listeners if STOP FADE is received."""
        self.update_listeners(True)

class DynaliteMultiDevice(DynaliteBaseDevice):
    """Representation of two Dynalite Presets as an on/off switch."""

    def __init__(self, num_devices, area, area_name, name, master_area, bridge):
        """Initialize the device."""
        super().__init__(area, area_name, name, master_area, bridge)
        self._devices = {}
        self._num_devices = num_devices

    def get_device(self, devnum):
        """Get the first or second device."""
        return self._devices.get(devnum)

    @property
    def available(self):
        """Return if dual device is available."""
        for i in range(1, self._num_devices + 1):
            if not self.get_device(i):
                return False
        return super().available

    def set_device(self, devnum, device):
        """Set one of the attached devices."""
        self._devices[devnum] = device
        device.add_listener(self.listener)
        if self.available:
            self._bridge.updateDevice(self)

    def listener(self, device, stop_fade):
        """Update the device since its internal devices changed."""
        self._bridge.updateDevice(self)



"""Dynalite Communications"""
import voluptuous as vol

from .const import *
from .dynalite_devices import DynaliteDevices
from dynalite_lib import (
    CONF_CHANNEL,
    CONF_AREA,
    CONF_PRESET,
    CONF_NODEFAULT,
    CONF_LOGLEVEL,
    CONF_FADE,
    CONF_DEFAULT,
    CONF_POLLTIMER,
    CONF_AUTO_DISCOVER,
)

DEFAULT_TEMPLATE_NAMES = [t for t in DEFAULT_TEMPLATES]


def numString(value):
    newValue = str(value)
    if newValue.isdigit():
        return newValue
    else:
        raise vol.Invalid("Not a string with numbers")


TEMPLATE_ROOM_SCHEMA = vol.Schema(
    {vol.Optional(CONF_ROOM_ON): numString, vol.Optional(CONF_ROOM_OFF): numString}
)

TEMPLATE_TRIGGER_SCHEMA = vol.Schema(
    {vol.Optional(CONF_TRIGGER): numString}
)

TEMPLATE_CHANNELCOVER_SCHEMA = vol.Schema(
    {
        vol.Optional(CONF_CHANNEL): numString,
        vol.Optional(CONF_CHANNEL_CLASS): str,
        vol.Optional(CONF_FACTOR): vol.Coerce(float),
        vol.Optional(CONF_TILT_PERCENTAGE): vol.Coerce(float),
    }
)

TEMPLATE_TIMECOVER_SCHEMA = vol.Schema(
    {
        vol.Optional(CONF_CHANNEL): numString,
        vol.Optional(CONF_CHANNEL_CLASS): str,
        vol.Optional(CONF_OPEN_PRESET): numString,
        vol.Optional(CONF_CLOSE_PRESET): numString,
        vol.Optional(CONF_STOP_PRESET): numString,
        vol.Optional(CONF_DURATION): vol.Coerce(float),
        vol.Optional(CONF_TILT_TIME): vol.Coerce(float),
    }
)

TEMPLATE_DATA_SCHEMA = vol.Any(
    TEMPLATE_ROOM_SCHEMA, TEMPLATE_TRIGGER_SCHEMA, TEMPLATE_CHANNELCOVER_SCHEMA, TEMPLATE_TIMECOVER_SCHEMA
)

PRESET_DATA_SCHEMA = vol.Schema(
    {
        vol.Optional(CONF_NAME): str,
        vol.Optional(CONF_FADE): vol.Coerce(float),
        vol.Optional(CONF_HIDDEN_ENTITY, default=False): vol.Coerce(bool),
    }
)

PRESET_SCHEMA = vol.Schema({numString: vol.Any(PRESET_DATA_SCHEMA, None)})


def check_channel_data_schema(conf): # XXX check if still relevant
    """Check that a channel config is valid."""
    if conf[CONF_CHANNEL_TYPE] != "cover":
        for param in [CONF_CHANNEL_CLASS, CONF_FACTOR, CONF_TILT_PERCENTAGE]:
            if param in conf:
                raise vol.Invalid(
                    "parameter " + param + " is only valid for 'cover' type channels"
                )
    return conf


CHANNEL_DATA_SCHEMA = vol.Schema(
    vol.All(
        {
            vol.Optional(CONF_NAME): str,
            vol.Optional(CONF_FADE): vol.Coerce(float),
            vol.Optional(CONF_CHANNEL_TYPE, default=DEFAULT_CHANNEL_TYPE): vol.Any(
                "light", "switch", "cover"
            ),
            vol.Optional(CONF_CHANNEL_CLASS): str,
            vol.Optional(CONF_HIDDEN_ENTITY, default=False): vol.Coerce(bool),
            vol.Optional(CONF_FACTOR): vol.Coerce(float),
            vol.Optional(CONF_TILT_PERCENTAGE): vol.Coerce(float),
            vol.Optional(CONF_PRESET): {numString: vol.Any(vol.Coerce(float), None)},
        },
        check_channel_data_schema,
    )
)

CHANNEL_SCHEMA = vol.Schema({numString: vol.Any(CHANNEL_DATA_SCHEMA, None)})


def check_area_data_schema(conf):
    """Verify that an area config is valid."""
    if CONF_TEMPLATE in conf and conf[CONF_TEMPLATE] not in DEFAULT_TEMPLATES:
        raise vol.Invalid(
            conf[CONF_TEMPLATE]
            + " is not a valid template name. Possible names are: "
            + str(DEFAULT_TEMPLATE_NAMES)
        )

    if CONF_TEMPLATE_OVERRIDE in conf and False:
        if CONF_TEMPLATE not in conf:
            raise vol.Invalid(
                CONF_TEMPLATE_OVERRIDE
                + " may only be present when "
                + CONF_TEMPLATE
                + " is defined"
            )
        template = conf[CONF_TEMPLATE]
        if template == CONF_ROOM:
            TEMPLATE_ROOM_SCHEMA(conf[CONF_TEMPLATE_OVERRIDE])
        elif template == CONF_TRIGGER:
            TEMPLATE_TRIGGER_SCHEMA(conf[CONF_TEMPLATE_OVERRIDE])
        else:
            raise vol.Invalid("Unknown template type " + template)
    return conf


AREA_DATA_SCHEMA = vol.Schema(
    vol.All(
        {
            vol.Required(CONF_NAME): str,
            vol.Optional(CONF_TEMPLATE): str,
            vol.Optional(CONF_TEMPLATE_OVERRIDE): TEMPLATE_DATA_SCHEMA,
            vol.Optional(CONF_FADE): vol.Coerce(float),
            vol.Optional(CONF_NODEFAULT): vol.Coerce(bool),
            vol.Optional(CONF_AREA_OVERRIDE): str,
            vol.Optional(CONF_PRESET): PRESET_SCHEMA,
            vol.Optional(CONF_CHANNEL): CHANNEL_SCHEMA,
        },
        check_area_data_schema,
    )
)

AREA_SCHEMA = vol.Schema({numString: vol.Any(AREA_DATA_SCHEMA, None)})

PLATFORM_DEFAULTS_SCHEMA = vol.Schema({vol.Optional(CONF_FADE): vol.Coerce(float)})


TEMPLATE_SCHEMA = vol.Schema({str: vol.Any(TEMPLATE_DATA_SCHEMA, None)})

BRIDGE_CONFIG_SCHEMA = vol.Schema(
    {
        vol.Optional(CONF_NAME, default=DEFAULT_NAME): str,
        vol.Required(CONF_HOST): str,
        vol.Optional(CONF_PORT, default=DEFAULT_PORT): int,
        vol.Optional(CONF_LOGLEVEL): str,
        vol.Optional(CONF_AUTO_DISCOVER, default=False): vol.Coerce(bool),
        vol.Optional(CONF_POLLTIMER, default=1.0): vol.Coerce(float),
        vol.Optional(CONF_AREA): AREA_SCHEMA,
        vol.Optional(CONF_DEFAULT): PLATFORM_DEFAULTS_SCHEMA,
        vol.Optional(CONF_PRESET): PRESET_SCHEMA,
        vol.Optional(CONF_TEMPLATE, default=DEFAULT_TEMPLATES): TEMPLATE_SCHEMA,
        vol.Optional(CONF_ACTIVE, default=False): vol.Coerce(bool),
    }
)



"""Class to create devices from a Dynalite hub."""

import copy
import asyncio

from .const import (
    LOGGER,
    CONF_TEMPLATE_OVERRIDE,
    DEFAULT_COVER_CHANNEL_CLASS,
    DEFAULT_COVER_FACTOR,
    CONF_TRIGGER,
    CONF_FACTOR,
    CONF_CHANNEL_TYPE,
    CONF_HIDDEN_ENTITY,
    CONF_TILT_PERCENTAGE,
    CONF_AREA_OVERRIDE,
    CONF_CHANNEL_CLASS,
    CONF_TEMPLATE,
    CONF_ROOM_ON,
    CONF_ROOM_OFF,
    DEFAULT_TEMPLATES,
    CONF_ROOM,
    DEFAULT_CHANNEL_TYPE,
    CONF_CHANNEL_COVER,
    CONF_NONE,
    CONF_TIME_COVER,
    CONF_OPEN_PRESET,
    CONF_CLOSE_PRESET,
    CONF_STOP_PRESET,
    CONF_DURATION,
    CONF_TILT_TIME,
)
from dynalite_lib import (
    CONF_CHANNEL,
    CONF_AREA,
    CONF_NAME,
    CONF_PRESET,
    CONF_NODEFAULT,
    EVENT_NEWPRESET,
    EVENT_NEWCHANNEL,
    EVENT_PRESET,
    EVENT_CHANNEL,
    EVENT_CONNECTED,
    EVENT_DISCONNECTED,
    EVENT_CONFIGURED,
    CONF_ACTION,
    CONF_ACTION_REPORT,
    CONF_ACTION_CMD,
    CONF_TRGT_LEVEL,
    CONF_ACT_LEVEL,
    CONF_ALL,
    Dynalite,
)

from .light import DynaliteChannelLightDevice
from .switch import (
    DynaliteChannelSwitchDevice,
    DynalitePresetSwitchDevice,
    DynaliteDualPresetSwitchDevice,
)
from .cover import DynaliteChannelCoverDevice, DynaliteChannelCoverWithTiltDevice, DynaliteTimeCoverDevice, DynaliteTimeCoverWithTiltDevice
from .dynalitebase import DynaliteBaseDevice


class BridgeError(Exception):
    """For errors in the Dynalite bridge."""

    def __init__(self, message):
        """Initialize the exception."""
        self.message = message


class DynaliteDevices:
    """Manages a single Dynalite bridge."""

    def __init__(self, config, loop=None, newDeviceFunc=None, updateDeviceFunc=None):
        """Initialize the system."""
        self.config = copy.deepcopy(config)
        self.loop = loop
        self.newDeviceFunc = newDeviceFunc
        self.updateDeviceFunc = updateDeviceFunc
        self.devices = []
        self.waiting_devices = []
        self.configured = False
        self.connected = False
        self.added_presets = {}
        self.added_channels = {}
        self.added_room_switches = {}
        self.added_time_covers = {}
        self.timer_active = False
        self.timer_callbacks = set()

    async def async_setup(self, tries=0):
        """Set up a Dynalite bridge based on host parameter in the config."""
        LOGGER.debug("bridge async_setup - %s", self.config)

        if not self.loop:
            self.loop = asyncio.get_running_loop()

        # insert the templates
        if CONF_TEMPLATE not in self.config:
            LOGGER.debug(CONF_TEMPLATE + " not in config - using defaults")
            self.config[CONF_TEMPLATE] = DEFAULT_TEMPLATES
        else:
            for template in DEFAULT_TEMPLATES:
                if template not in self.config[CONF_TEMPLATE]:
                    LOGGER.debug(
                        "%s not in " + CONF_TEMPLATE + " using default", template
                    )
                    self.config[CONF_TEMPLATE][template] = DEFAULT_TEMPLATES[template]
                else:
                    for param in DEFAULT_TEMPLATES[template]:
                        if param not in self.config[CONF_TEMPLATE][template]:
                            self.config[CONF_TEMPLATE][template][
                                param
                            ] = DEFAULT_TEMPLATES[template][param]
        # add the entities implicitly defined by templates
        for curArea in self.config[CONF_AREA]:
            areaConfig = self.config[CONF_AREA][curArea]
            if CONF_TEMPLATE in areaConfig:
                template = self.config[CONF_AREA][curArea][CONF_TEMPLATE]
                template_params = self.getTemplateParams(curArea)
                if template == CONF_ROOM:
                    areaConfig[CONF_NODEFAULT] = True
                    for conf in [CONF_ROOM_ON, CONF_ROOM_OFF]:
                        self.ensurePresetInConfig(curArea, template_params[conf])
                        areaConfig[conf] = template_params[conf]
                elif template == CONF_TRIGGER:
                    areaConfig[CONF_NODEFAULT] = True
                    self.ensurePresetInConfig(curArea, template_params[CONF_TRIGGER], False, areaConfig[CONF_NAME])
                    areaConfig[CONF_TRIGGER] = template_params[CONF_TRIGGER]
                elif template == CONF_CHANNEL_COVER:
                    areaConfig[CONF_NODEFAULT] = True
                    curChannel = template_params[CONF_CHANNEL]
                    self.ensureChannelInConfig(curArea, curChannel, False, areaConfig[CONF_NAME])
                    areaConfig[CONF_CHANNEL][str(curChannel)][CONF_CHANNEL_TYPE] = "cover"
                    for conf in [CONF_CHANNEL_CLASS, CONF_FACTOR, CONF_TILT_PERCENTAGE]:
                        areaConfig[CONF_CHANNEL][str(curChannel)][conf] = template_params[conf]
                elif template == CONF_TIME_COVER:
                    areaConfig[CONF_NODEFAULT] = True
                    curChannel = template_params[CONF_CHANNEL_COVER]
                    if int(curChannel) > 0:
                        areaConfig[CONF_CHANNEL_COVER] = curChannel
                        self.ensureChannelInConfig(curArea, curChannel)
                    for conf in [CONF_OPEN_PRESET, CONF_CLOSE_PRESET, CONF_STOP_PRESET]:
                        self.ensurePresetInConfig(curArea, template_params[conf])
                        areaConfig[conf] = template_params[conf]
                    for conf in [CONF_CHANNEL_CLASS, CONF_DURATION, CONF_TILT_TIME]:
                        areaConfig[conf] = template_params[conf]
        LOGGER.debug("bridge async_setup (after templates) - %s" % self.config)

        # Configure the dynalite object
        self._dynalite = Dynalite(config=self.config, loop=self.loop)
        eventHandler = self._dynalite.addListener(listenerFunction=self.handleEvent)
        eventHandler.monitorEvent("*")
        newPresetHandler = self._dynalite.addListener(
            listenerFunction=self.handleNewPreset
        )
        newPresetHandler.monitorEvent(EVENT_NEWPRESET)
        presetChangeHandler = self._dynalite.addListener(
            listenerFunction=self.handlePresetChange
        )
        presetChangeHandler.monitorEvent(EVENT_PRESET)
        newChannelHandler = self._dynalite.addListener(
            listenerFunction=self.handleNewChannel
        )
        newChannelHandler.monitorEvent(EVENT_NEWCHANNEL)
        channelChangeHandler = self._dynalite.addListener(
            listenerFunction=self.handleChannelChange
        )
        channelChangeHandler.monitorEvent(EVENT_CHANNEL)
        self._dynalite.start()

        # register the rooms (switches on presets 1/4)
        self.registerRooms()

        # register the time covers
        self.registerTimeCovers()

        return True

    def ensurePresetInConfig(self, area, preset, hidden=True, name=None):
        areaConfig = self.config[CONF_AREA][area]
        if CONF_PRESET not in areaConfig:
            areaConfig[CONF_PRESET] = {}
        if str(preset) not in areaConfig[CONF_PRESET]:
            curConf = {CONF_HIDDEN_ENTITY: hidden}
            if name:
                curConf[CONF_NAME] = name
            areaConfig[CONF_PRESET][str(preset)] = curConf

    def ensureChannelInConfig(self, area, channel, hidden=True, name=None):
        areaConfig = self.config[CONF_AREA][area]
        if CONF_CHANNEL not in areaConfig:
            areaConfig[CONF_CHANNEL] = {}
        if str(channel) not in areaConfig[CONF_CHANNEL]:
            curConf = {CONF_HIDDEN_ENTITY: hidden}
            if name:
                curConf[CONF_NAME] = name
            areaConfig[CONF_CHANNEL][str(channel)] = curConf

    def registerRooms(self):
        """Register the room switches from two normal presets each."""
        for curArea, areaConfig in self.config[CONF_AREA].items():
            if CONF_TEMPLATE in areaConfig and areaConfig[CONF_TEMPLATE] == CONF_ROOM:
                newDevice = DynaliteDualPresetSwitchDevice(
                    curArea,
                    areaConfig[CONF_NAME],
                    areaConfig[CONF_NAME],
                    self.getMasterArea(curArea),
                    self,
                )
                self.added_room_switches[int(curArea)] = newDevice
                self.setPresetIfReady(curArea, areaConfig[CONF_ROOM_ON], 1, newDevice)
                self.setPresetIfReady(curArea, areaConfig[CONF_ROOM_OFF], 2, newDevice)
                self.registerNewDevice("switch", newDevice, False)

    def registerTimeCovers(self):
        """Register the time covers from three presets and a channel each."""
        for curArea, areaConfig in self.config[CONF_AREA].items():
            if CONF_TEMPLATE in areaConfig and areaConfig[CONF_TEMPLATE] == CONF_TIME_COVER:
                if areaConfig[CONF_TILT_TIME] == 0:
                    newDevice = DynaliteTimeCoverDevice(
                        curArea,
                        areaConfig[CONF_NAME],
                        areaConfig[CONF_NAME],
                        areaConfig[CONF_DURATION],
                        areaConfig[CONF_CHANNEL_CLASS],
                        self.getMasterArea(curArea),
                        self,
                    )
                else:
                    newDevice = DynaliteTimeCoverWithTiltDevice(
                        curArea,
                        areaConfig[CONF_NAME],
                        areaConfig[CONF_NAME],
                        areaConfig[CONF_DURATION],
                        areaConfig[CONF_CHANNEL_CLASS],
                        areaConfig[CONF_TILT_TIME],
                        self.getMasterArea(curArea),
                        self,
                    )
                self.added_time_covers[int(curArea)] = newDevice
                self.setPresetIfReady(curArea, areaConfig[CONF_OPEN_PRESET], 1, newDevice)
                self.setPresetIfReady(curArea, areaConfig[CONF_CLOSE_PRESET], 2, newDevice)
                self.setPresetIfReady(curArea, areaConfig[CONF_STOP_PRESET], 3, newDevice)
                if CONF_CHANNEL_COVER in areaConfig:
                    self.setChannelIfReady(curArea, areaConfig[CONF_CHANNEL_COVER], 4, newDevice)
                else:
                    # No channel attached, only presets.
                    dummyDevice = DynaliteBaseDevice(
                        curArea,
                        areaConfig[CONF_NAME],
                        areaConfig[CONF_NAME],
                        self.getMasterArea(curArea),
                        self,
                    )
                    newDevice.set_device(4, dummyDevice)
                self.registerNewDevice("cover", newDevice, False)

    def registerNewDevice(self, category, device, hidden):
        """Register a new device and group all the ones prior to CONFIGURED event together."""
        self.devices.append(device)
        # after initial configuration, every new device gets sent on its own. The initial ones are bunched together
        if not hidden:
            if self.configured:  
                if self.newDeviceFunc:
                    self.newDeviceFunc([device])
            else:  # send all the devices together when configured
                self.waiting_devices.append(device)

    @property
    def available(self):
        """Return whether bridge is available."""
        return self.connected

    def updateDevice(self, device):
        """Update one or more devices."""
        if self.updateDeviceFunc:
            self.updateDeviceFunc(device)

    def getTemplateIndex(self, area, template, conf):
        """Get a specific index from a specific template in an area."""
        # should always be defined either by the user or by the defaults
        # first see if it is in the area
        try:
            return self.config[CONF_AREA][str(area)][conf]
        except KeyError:
            pass
        # next try templateoverride
        try:
            return self.config[CONF_AREA][str(area)][CONF_TEMPLATE_OVERRIDE][conf]
        except KeyError:
            pass
        # next try the template in the config
        # should always be defined either by the user or by the defaults
        my_template = self.config[CONF_TEMPLATE][template]  
        if conf in my_template:
            return my_template[conf]
        # not found
        return None

    def getTemplateParams(self, area):
        """Extract all the parameters of a template."""
        result = {}
        area_config = self.config[CONF_AREA][str(area)]
        template = area_config.get(CONF_TEMPLATE)
        if not template:
            return result
        for conf in self.config[CONF_TEMPLATE][template]:
            value = self.getTemplateIndex(area, template, conf)
            result[conf] = value
        return result

    def setPresetIfReady(self, area, preset, deviceNum, multiDevice):
        """Try to set a preset of a multi device if it was already registered."""
        try:
            device = self.added_presets[int(area)][int(preset)]
            multiDevice.set_device(deviceNum, device)
        except KeyError:
            pass

    def setChannelIfReady(self, area, channel, deviceNum, multiDevice):
        """Try to set a preset of a multi device if it was already registered."""
        try:
            device = self.added_channels[int(area)][int(channel)]
            multiDevice.set_device(deviceNum, device)
        except KeyError:
            pass

    def handleEvent(self, event=None, dynalite=None):
        """Handle all events."""
        LOGGER.debug("handleEvent - type=%s event=%s" % (event.eventType, event.data))
        if event.eventType == EVENT_CONNECTED:
            LOGGER.debug("received CONNECTED message")
            self.connected = True
            self.updateDevice(CONF_ALL)
        elif event.eventType == EVENT_DISCONNECTED:
            LOGGER.debug("received DISCONNECTED message")
            self.connected = False
            self.updateDevice(CONF_ALL)
        elif event.eventType == EVENT_CONFIGURED:
            LOGGER.debug("received CONFIGURED message")
            self.configured = True
            if self.newDeviceFunc and self.waiting_devices:
                self.newDeviceFunc(self.waiting_devices)
                self.waiting_devices = []
        return

    def getMasterArea(self, area):
        """Get the master area when combining entities from different Dynet areas to the same area."""
        if str(area) not in self.config[CONF_AREA]:
            LOGGER.error("getMasterArea - we should not get here")
            raise BridgeError("getMasterArea - area " + str(area) + "is not in config")
        areaConfig = self.config[CONF_AREA][str(area)]
        masterArea = areaConfig[CONF_NAME]
        if CONF_AREA_OVERRIDE in areaConfig:
            overrideArea = areaConfig[CONF_AREA_OVERRIDE]
            masterArea = overrideArea if overrideArea.lower() != CONF_NONE else ""
        return masterArea

    def handleNewPreset(self, event=None, dynalite=None):
        """Register a new preset."""
        LOGGER.debug("handleNewPreset - event=%s", event.data)
        if not hasattr(event, "data"):
            return
        if CONF_AREA not in event.data:
            return
        curArea = event.data[CONF_AREA]
        if CONF_PRESET not in event.data:
            return
        curPreset = event.data[CONF_PRESET]

        if str(curArea) not in self.config[CONF_AREA]:
            LOGGER.debug("adding area " + str(curArea) + " that is not in config")
            self.config[CONF_AREA][str(curArea)] = {CONF_NAME: "Area " + str(curArea)}
        areaConfig = self.config[CONF_AREA][str(curArea)]

        try:
            # If the name is explicitly defined, use it
            curName = areaConfig[CONF_PRESET][str(curPreset)][CONF_NAME]  
        except KeyError:
            presetName = "Preset " + str(curPreset)
            if CONF_NODEFAULT not in areaConfig or not areaConfig[CONF_NODEFAULT]:
                try:
                    presetName = self.config[CONF_PRESET][str(curPreset)][CONF_NAME]
                except KeyError:
                    pass
            # If not explicitly defined, use "areaname presetname"
            curName = areaConfig[CONF_NAME] + " " + presetName  
        curDevice = self._dynalite.devices[CONF_AREA][curArea].preset[curPreset]
        newDevice = DynalitePresetSwitchDevice(
            curArea,
            areaConfig[CONF_NAME],
            curPreset,
            curName,
            self.getMasterArea(curArea),
            self,
            curDevice,
        )

        try:
            hidden = areaConfig[CONF_PRESET][str(curPreset)][CONF_HIDDEN_ENTITY]
        except KeyError:
            hidden = False

        try:
            # templates may make some elements hidden or register the preset
            template = areaConfig[CONF_TEMPLATE]  
            if template == CONF_ROOM:
                # in a template room, the presets will all be in the room switch
                hidden = True  
                # if it is not there yet, it will be added when the room switch will be created
                if int(curArea) in self.added_room_switches:  
                    multiDevice = self.added_room_switches[int(curArea)]
                    if int(curPreset) == int(areaConfig[CONF_ROOM_ON]):
                        multiDevice.set_device(1, newDevice)
                    if int(curPreset) == int(areaConfig[CONF_ROOM_OFF]):
                        multiDevice.set_device(2, newDevice)
            elif template == CONF_TRIGGER:
                if int(curPreset) != areaConfig[CONF_TRIGGER]:
                    hidden = True
            elif template in [CONF_HIDDEN_ENTITY, CONF_CHANNEL_COVER]:
                hidden = True
            elif template == CONF_TIME_COVER:
                # in a template room, the presets will all be in the time cover
                hidden = True  
                # if it is not there yet, it will be added when the time cover will be created
                if int(curArea) in self.added_time_covers:
                    multiDevice = self.added_time_covers[int(curArea)]
                    if int(curPreset) == int(areaConfig[CONF_OPEN_PRESET]):
                        multiDevice.set_device(1, newDevice)
                    if int(curPreset) == int(areaConfig[CONF_CLOSE_PRESET]):
                        multiDevice.set_device(2, newDevice)
                    if int(curPreset) == int(areaConfig[CONF_STOP_PRESET]):
                        multiDevice.set_device(3, newDevice)
            else:
                LOGGER.error(
                    "Unknown template "
                    + template
                    + ". Should have been caught in config_validation"
                )
        except KeyError:
            pass

        self.registerNewDevice("switch", newDevice, hidden)
        if curArea not in self.added_presets:
            self.added_presets[curArea] = {}
        self.added_presets[curArea][curPreset] = newDevice
        LOGGER.debug(
            "Creating Dynalite preset area=%s preset=%s name=%s hidden=%s", curArea, curPreset, curName, hidden
        )

    def handlePresetChange(self, event=None, dynalite=None):
        """Change the selected preset."""
        LOGGER.debug("handlePresetChange - event=%s" % event.data)
        if not hasattr(event, "data"):
            return
        if CONF_AREA not in event.data:
            return
        curArea = event.data[CONF_AREA]
        if CONF_PRESET not in event.data:
            return

        # Update all the preset devices
        if int(curArea) in self.added_presets:
            for curPresetInArea in self.added_presets[int(curArea)]:
                self.updateDevice(self.added_presets[int(curArea)][curPresetInArea])

    def handleNewChannel(self, event=None, dynalite=None):
        """Register a new channel."""
        LOGGER.debug("handleNewChannel - event=%s" % event.data)
        if not hasattr(event, "data"):
            return
        if CONF_AREA not in event.data:
            return
        curArea = event.data[CONF_AREA]
        if CONF_CHANNEL not in event.data:
            return
        curChannel = event.data[CONF_CHANNEL]

        if str(curArea) not in self.config[CONF_AREA]:
            LOGGER.debug("adding area " + str(curArea) + " that is not in config")
            self.config[CONF_AREA][str(curArea)] = {CONF_NAME: "Area " + str(curArea)}
        areaConfig = self.config[CONF_AREA][str(curArea)]

        try:
            # If the name is explicitly defined, use it
            curName = areaConfig[CONF_CHANNEL][str(curChannel)][CONF_NAME]  
        except (KeyError, TypeError):
            # If not explicitly defined, use "areaname Channel X"
            curName = areaConfig[CONF_NAME] + " Channel " + str(curChannel)
        curDevice = self._dynalite.devices[CONF_AREA][curArea].channel[curChannel]
        try:
            channelConfig = areaConfig[CONF_CHANNEL][str(curChannel)]
        except KeyError:
            channelConfig = None
        LOGGER.debug("handleNewChannel - channelConfig=%s" % channelConfig)
        channelType = (
            channelConfig[CONF_CHANNEL_TYPE].lower()
            if channelConfig and CONF_CHANNEL_TYPE in channelConfig
            else DEFAULT_CHANNEL_TYPE
        )
        hassArea = self.getMasterArea(curArea)
        hidden = (channelConfig and CONF_HIDDEN_ENTITY in channelConfig and channelConfig[CONF_HIDDEN_ENTITY]) or \
                 (self.config[CONF_AREA][str(curArea)].get(CONF_TEMPLATE) == CONF_HIDDEN_ENTITY)
        if channelType == "light":
            newDevice = DynaliteChannelLightDevice(
                curArea,
                areaConfig[CONF_NAME],
                curChannel,
                curName,
                channelType,
                hassArea,
                self,
                curDevice,
            )
            self.registerNewDevice("light", newDevice, hidden)
        elif channelType == "switch":
            newDevice = DynaliteChannelSwitchDevice(
                curArea,
                areaConfig[CONF_NAME],
                curChannel,
                curName,
                channelType,
                hassArea,
                self,
                curDevice,
            )
            self.registerNewDevice("switch", newDevice, hidden)
        elif channelType == "cover":
            factor = (
                channelConfig[CONF_FACTOR]
                if CONF_FACTOR in channelConfig
                else DEFAULT_COVER_FACTOR
            )
            deviceClass = (
                channelConfig[CONF_CHANNEL_CLASS]
                if CONF_CHANNEL_CLASS in channelConfig
                else DEFAULT_COVER_CHANNEL_CLASS
            )
            if CONF_TILT_PERCENTAGE in channelConfig and channelConfig[CONF_TILTPERCENTAGE] != 0:
                newDevice = DynaliteChannelCoverWithTiltDevice(
                    curArea,
                    areaConfig[CONF_NAME],
                    curChannel,
                    curName,
                    channelType,
                    deviceClass,
                    factor,
                    channelConfig[CONF_TILT_PERCENTAGE],
                    hassArea,
                    self,
                    curDevice,
                )
            else:
                newDevice = DynaliteChannelCoverDevice(
                    curArea,
                    areaConfig[CONF_NAME],
                    curChannel,
                    curName,
                    channelType,
                    deviceClass,
                    factor,
                    hassArea,
                    self,
                    curDevice,
                )
            self.registerNewDevice("cover", newDevice, hidden)
        else:
            LOGGER.info("unknown chnanel type %s - ignoring", channelType)
            return
        if curArea not in self.added_channels:
            self.added_channels[curArea] = {}
        self.added_channels[curArea][curChannel] = newDevice
        LOGGER.debug("Creating Dynalite channel area=%s channel=%s name=%s", curArea, curChannel, curName)
        # if it is a channel from a timecover, register it
        try:
            template = areaConfig[CONF_TEMPLATE]  
            if template == CONF_TIME_COVER:
                # in a template room, the channels will all be in the time cover
                hidden = True  
                # if it is not there yet, it will be added when the time cover will be created
                if int(curArea) in self.added_time_covers:  
                    multiDevice = self.added_time_covers[int(curArea)]
                    if int(curChannel) == int(areaConfig[CONF_CHANNEL_COVER]):
                        multiDevice.set_device(4, newDevice)
        except KeyError:
            pass

    def handleChannelChange(self, event=None, dynalite=None):
        """Change the level of a channel."""
        LOGGER.debug("handleChannelChange - event=%s" % event.data)
        LOGGER.debug("handleChannelChange called event = %s" % event.msg)
        if not hasattr(event, "data"):
            return
        if CONF_AREA not in event.data:
            return
        curArea = event.data[CONF_AREA]
        if CONF_CHANNEL not in event.data:
            return
        curChannel = event.data[CONF_CHANNEL]

        action = event.data[CONF_ACTION]
        if action == CONF_ACTION_REPORT:
            actual_level = (255 - event.data[CONF_ACT_LEVEL]) / 254
            target_level = (255 - event.data[CONF_TRGT_LEVEL]) / 254
        elif action == CONF_ACTION_CMD:
            if CONF_TRGT_LEVEL in event.data:
                target_level = (255 - event.data[CONF_TRGT_LEVEL]) / 254
                # when there is only a "set channel level" command, assume that this is both the actual and the target
                actual_level = target_level
            else: # stop fade command
                try:
                    if curChannel == CONF_ALL:
                        for channel in self.added_channels[int(curArea)]:
                            channelToSet = self.added_channels[int(curArea)][channel]
                            channelToSet.stop_fade()
                            self.updateDevice(channelToSet)  
                    else:
                        channelToSet = self.added_channels[int(curArea)][int(curChannel)]
                        channelToSet.stop_fade()
                        self.updateDevice(channelToSet)  
                except KeyError:
                    pass
                return
        else:
            LOGGER.error("unknown action for channel change %s", action)
            return
        try:
            channelToSet = self.added_channels[int(curArea)][int(curChannel)]
            channelToSet.update_level(actual_level, target_level)
            # to only call if it was already added to ha
            self.updateDevice(channelToSet)  
        except KeyError:
            pass
            
    def add_timer_listener(self, callback_func):
        # LOGGER.debug("XXX add timer %s", callback_func)
        self.timer_callbacks.add(callback_func)
        if not self.timer_active:
            self.loop.call_later(1, self.timer_func)
            self.timer_active = True
            
    def remove_timer_listener(self, callback_func):
        # LOGGER.debug("XXX remove timer %s", callback_func)
        self.timer_callbacks.discard(callback_func)
        
    def timer_func(self):
        if self.timer_callbacks:
            for callback in self.timer_callbacks:
                self.loop.call_soon(callback)
            self.loop.call_later(1, self.timer_func)
        else:
            self.timer_active = False
        



"""Support for the Dynalite channels as covers."""
from .const import ATTR_POSITION, ATTR_TILT_POSITION, LOGGER

from .dynalitebase import DynaliteChannelBaseDevice, DynaliteMultiDevice
import asyncio

class DynaliteChannelCoverDevice(DynaliteChannelBaseDevice):
    """Representation of a Dynalite Channel as a Home Assistant Cover."""

    def __init__(
        self,
        area,
        area_name,
        channel,
        name,
        type,
        device_class,
        cover_factor,
        master_area,
        bridge,
        device,
    ):
        """Initialize the cover."""
        self._device_class = device_class
        self._cover_factor = cover_factor
        self._actual_level = 0
        self._target_level = 0
        self._current_position = 0
        super().__init__(
            area, area_name, channel, name, type, master_area, bridge, device
        )

    @property
    def category(self):
        """Return the category of the entity: light, switch, or cover."""
        return "cover"

    @property
    def has_tilt(self):
        """Return whether cover supports tilt."""
        return False

    @property
    def device_class(self):
        """Return the class of the cover."""
        return self._device_class

    def update_level(self, actual_level, target_level):
        """Update the current level."""
        prev_actual_level = self._actual_level
        self._actual_level = actual_level
        self._target_level = target_level
        level_diff = actual_level - prev_actual_level
        factored_diff = level_diff / self._cover_factor
        self._current_position = min(1, max(0, self._current_position + factored_diff))
        if self._current_position > 0.99999:
            self._current_position = 1
        if self._current_position < 0.00001:
            self._current_position = 0
        if getattr(self, "update_tilt", False):
            self.update_tilt(factored_diff)

    @property
    def current_cover_position(self):
        """Return the position of the cover from 0 to 100."""
        return int(self._current_position * 100)

    @property
    def is_opening(self):
        """Return whether cover is currently opening."""
        return self._current_position < 1 and self._target_level > self._actual_level

    @property
    def is_closing(self):
        """Return whether cover is currently closing."""
        return self._current_position > 0 and self._target_level < self._actual_level

    @property
    def is_closed(self):
        """Return whether cover is closed."""
        return self._current_position == 0

    async def async_open_cover(self, **kwargs):
        """Open the cover."""
        self._device.turnOn()

    async def async_close_cover(self, **kwargs):
        """Close the cover."""
        self._device.turnOff()

    async def async_set_cover_position(self, **kwargs):
        """Set the cover to a specific position."""
        target_position = kwargs[ATTR_POSITION] / 100
        position_diff = target_position - self._current_position
        level_diff = position_diff * self._cover_factor
        target_level = min(1, max(0, self._actual_level + level_diff))
        self._device.turnOn(brightness=target_level)

    async def async_stop_cover(self, **kwargs):
        """Stop the cover."""
        self._device.stopFade()


class DynaliteChannelCoverWithTiltDevice(DynaliteChannelCoverDevice):
    """Representation of a Dynalite Channel as a Home Assistant Cover that uses up and down for tilt."""

    def __init__(
        self,
        area,
        area_name,
        channel,
        name,
        type,
        device_class,
        cover_factor,
        tilt_percentage,
        master_area,
        bridge,
        device,
    ):
        """Initialize the cover."""
        super().__init__(
            area,
            area_name,
            channel,
            name,
            type,
            device_class,
            cover_factor,
            master_area,
            bridge,
            device,
        )
        self._tilt_percentage = tilt_percentage
        self._current_tilt = 0

    @property
    def has_tilt(self):
        """Return whether cover supports tilt."""
        return True

    def update_tilt(self, diff):
        """Update the current tilt based on diff and tilt_percentage."""
        tilt_diff = diff / self._tilt_percentage
        self._current_tilt = max(0, min(1, self._current_tilt + tilt_diff))

    @property
    def current_cover_tilt_position(self):
        """Return the current cover tilt."""
        return int(self._current_tilt * 100)

    async def apply_tilt_diff(self, tilt_diff):
        """Move the cover up or down based on a diff."""
        position_diff = tilt_diff * self._tilt_percentage
        target_position = int(
            100 * max(0, min(1, self._current_position + position_diff))
        )
        await self.async_set_cover_position(position=target_position)

    async def async_open_cover_tilt(self, **kwargs):
        """Open the cover tilt."""
        if self._current_tilt == 1:
            return
        else:
            await self.apply_tilt_diff(1 - self._current_tilt)

    async def async_close_cover_tilt(self, **kwargs):
        """Close the cover tilt."""
        if self._current_tilt == 0:
            return
        else:
            await self.apply_tilt_diff(0 - self._current_tilt)

    async def async_set_cover_tilt_position(self, **kwargs):
        """Set the cover tilt position."""
        target_position = kwargs[ATTR_TILT_POSITION] / 100
        await self.apply_tilt_diff(target_position - self._current_tilt)

    async def async_stop_cover_tilt(self, **kwargs):
        """Stop cover tilt."""
        await self.async_stop_cover()

class DynaliteTimeCoverDevice(DynaliteMultiDevice):
    """Representation of a Dynalite Channel as a Home Assistant Cover."""

    def __init__(
        self,
        area,
        area_name,
        name,
        duration,
        device_class,
        master_area,
        bridge,
    ):
        """Initialize the cover."""
        self._device_class = device_class
        self._current_position = 0
        self._duration = duration
        self._direction = "stop"
        super().__init__(4, area, area_name, name, master_area, bridge)

    @property
    def category(self):
        """Return the category of the entity: light, switch, or cover."""
        return "cover"

    @property
    def unique_id(self):
        """Return the ID of this room switch."""
        return "dynalite_area_" + str(self._area) + "_time_cover"

    @property
    def has_tilt(self):
        """Return whether cover supports tilt."""
        return False

    @property
    def device_class(self):
        """Return the class of the cover."""
        return self._device_class

    def update_level(self, actual_level, target_level):
        """Update the current level."""
        if actual_level == target_level:
            self._direction = "stop"
            self._bridge.remove_timer_listener(self.timer_callback)
        else:
            self._direction = "open" if target_level > actual_level else "close"
            self._bridge.add_timer_listener(self.timer_callback)
        self._bridge.updateDevice(self)
            
    def timer_callback(self):
        if self._direction == "open":
            self._current_position += 1.0 / self._duration
            if self._current_position >= 1.0:
                self._current_position = 1.0
                self._direction = "stop"
                self._bridge.remove_timer_listener(self.timer_callback)
        elif self._direction == "close":    
            self._current_position -= 1.0 / self._duration
            if self._current_position <= 0.0:
                self._current_position = 0.0
                self._direction = "stop"
                self._bridge.remove_timer_listener(self.timer_callback)
        elif self._direction == "stop":
            self._bridge.remove_timer_listener(self.timer_callback)

        if getattr(self, "update_tilt", False):
            self.update_tilt()

        self._bridge.updateDevice(self)

    @property
    def current_cover_position(self):
        """Return the position of the cover from 0 to 100."""
        return int(self._current_position * 100)

    @property
    def is_opening(self):
        """Return whether cover is currently opening."""
        return self._direction == "open"

    @property
    def is_closing(self):
        """Return whether cover is currently closing."""
        return self._direction == "close"

    @property
    def is_closed(self):
        """Return whether cover is closed."""
        return self._current_position == 0

    async def async_open_cover(self, **kwargs):
        """Open the cover."""
        await self.get_device(1).async_turn_on()
        self.update_level(self._current_position, 1.0)

    async def async_close_cover(self, **kwargs):
        """Close the cover."""
        await self.get_device(2).async_turn_on()
        self.update_level(self._current_position, 0.0)

    async def async_set_cover_position(self, **kwargs):
        """Set the cover to a specific position."""
        target_position = kwargs[ATTR_POSITION] / 100
        position_diff = target_position - self._current_position
        if position_diff > 0:
            await self.async_open_cover()
            while self._current_position < target_position and self._direction == "open":
                await asyncio.sleep(1)
            if self._direction == "open":
                await self.async_stop_cover()
                await asyncio.sleep(1) # doing twice for safety
                await self.async_stop_cover()
        elif position_diff < 0:
            await self.async_close_cover()
            while self._current_position > target_position and self._direction == "close":
                await asyncio.sleep(1)
            if self._direction == "close":
                await self.async_stop_cover()
                await asyncio.sleep(1) # doing twice for safety
                await self.async_stop_cover()
        else:
            await self.async_stop_cover()
        
    async def async_stop_cover(self, **kwargs):
        """Stop the cover."""
        await self.get_device(3).async_turn_on()
        self.update_level(self._current_position, self._current_position)
        
    def listener(self, device, stop_fade):
        if device == self.get_device(1):
            if device.level > 0:
                self.update_level(self._current_position, 1.0)
        elif device == self.get_device(2):
            if device.level > 0:
                self.update_level(self._current_position, 0.0)
        elif device == self.get_device(3):
            if device.level > 0:
                self.update_level(self._current_position, self._current_position)
        elif device == self.get_device(4):
            if stop_fade or device.direction == "stop":
                self.update_level(self._current_position, self._current_position)
            elif device.direction == "open":
                self.update_level(self._current_position, 1.0)
            else:
                self.update_level(self._current_position, 0.0)
                
        else:
            LOGGER.error("listener received update from unknown device")
        super().listener(device, stop_fade)

class DynaliteTimeCoverWithTiltDevice(DynaliteTimeCoverDevice):
    """Representation of a Dynalite Channel as a Home Assistant Cover that uses up and down for tilt."""

    def __init__(
        self,
        area,
        area_name,
        name,
        duration,
        device_class,
        tilt_duration,
        master_area,
        bridge,
    ):
        """Initialize the cover."""
        super().__init__(
            area,
            area_name,
            name,
            duration,
            device_class,
            master_area,
            bridge,
        )
        self._tilt_duration = tilt_duration
        self._current_tilt = 0

    @property
    def has_tilt(self):
        """Return whether cover supports tilt."""
        return True

    def update_tilt(self):
        """Update the current tilt based on diff and tilt_percentage."""
        if self._direction == "open":
            mult = 1
        elif self._direction == "close":
            mult = -1
        else:
            LOGGER.error("update_tilt called with invalid direction %s", self._direction)
            return
        tilt_diff = mult / self._tilt_duration
        self._current_tilt = max(0, min(1, self._current_tilt + tilt_diff))

    @property
    def current_cover_tilt_position(self):
        """Return the current cover tilt."""
        return int(self._current_tilt * 100)

    async def apply_tilt_diff(self, tilt_diff):
        """Move the cover up or down based on a diff."""
        factor = self._tilt_duration / self._duration
        position_diff = tilt_diff * factor
        target_position = int(
            100 * max(0, min(1, self._current_position + position_diff))
        )
        await self.async_set_cover_position(position=target_position)

    async def async_open_cover_tilt(self, **kwargs):
        """Open the cover tilt."""
        if self._current_tilt == 1:
            return
        else:
            await self.apply_tilt_diff(1 - self._current_tilt)

    async def async_close_cover_tilt(self, **kwargs):
        """Close the cover tilt."""
        if self._current_tilt == 0:
            return
        else:
            await self.apply_tilt_diff(0 - self._current_tilt)

    async def async_set_cover_tilt_position(self, **kwargs):
        """Set the cover tilt position."""
        target_position = kwargs[ATTR_TILT_POSITION] / 100
        await self.apply_tilt_diff(target_position - self._current_tilt)

    async def async_stop_cover_tilt(self, **kwargs):
        """Stop cover tilt."""
        await self.async_stop_cover()
